"""Local models repository tests."""
from pathlib import Path

import pytest

from tests.test_tools import create_dir
from thipster.engine.resource_model import ResourceModel
from thipster.repository import LocalRepo
from thipster.repository.exceptions import ModelNotFoundError

test_bucket = 'test/bucket'


def __setup_local(models: dict[str, str]):
    path_input = 'test'
    return create_dir(
        path_input,
        models,
    )


def test_get_bucket():
    """Test get bucket model."""
    _destroy_dir = __setup_local({
        'bucket.json':
            """
{
    "dependencies": {},
    "internalObjects": {},
    "attributes":{
        "region": {
            "optional": true,
            "default": "euw",
            "cdk_key": "region",
            "var_type": "str"
        }
    },
    "cdk_name_key": "name",

    "cdk_provider":"cdktf_cdktf_provider_google",
    "cdk_module":"storage_bucket",
    "cdk_class":"StorageBucket"
}
""",
    })
    resources = [test_bucket]
    repo = LocalRepo(Path.cwd().as_posix())

    models = repo.get(resources)
    _destroy_dir()

    assert isinstance(models, dict)
    assert len(models) == 1

    assert test_bucket in models
    assert isinstance(models[test_bucket], ResourceModel)

    bucket = models[test_bucket]

    assert len(bucket.attributes) == 1
    assert len(bucket.dependencies) == 0
    assert bucket.resource_type == test_bucket


def test_get_bucket_with_cors():
    """Test get bucket model with 'cors' internal object."""
    _destroy_dir = __setup_local({
        'bucket.json':
            """
{
    "dependencies": {},
    "internalObjects": {
        "cors": {
            "resource" : "test/bucket_cors",
            "var_type": "list[StorageBucketCors]",
            "default": {}
        }
    },
    "attributes":{
        "region": {
            "optional": true,
            "default": "euw",
            "cdk_key": "region",
            "var_type": "str"
        }
    },
    "cdk_name_key": "name",

    "cdk_provider":"cdktf_cdktf_provider_google",
    "cdk_module":"storage_bucket",
    "cdk_class":"StorageBucket"
}
""",
        'bucket_cors.json':
            """
{
    "dependencies": {},
    "internalObjects": {},
    "attributes":{
        "origin": {
            "optional": true,
            "cdk_key": "origin",
            "var_type": "list[str]"
        },
        "method": {
            "optional": true,
            "cdk_key": "method",
            "var_type": "list[str]"
        },
        "responseHeader": {
            "optional": true,
            "cdk_key": "response_header",
            "var_type": "list[str]"
        },
        "maxAge": {
            "optional": true,
            "cdk_key": "max_age_seconds",
            "var_type": "int"
        }
    },

    "cdk_provider":"cdktf_cdktf_provider_google",
    "cdk_module":"storage_bucket",
    "cdk_class":"StorageBucketCors"
}
""",
    })
    resources = [test_bucket]
    repo = LocalRepo(Path.cwd().as_posix())

    models = repo.get(resources)
    _destroy_dir()

    assert isinstance(models, dict)
    assert len(models) == 2

    assert test_bucket in models
    assert isinstance(models[test_bucket], ResourceModel)

    bucket = models[test_bucket]

    assert len(bucket.attributes) == 1
    assert len(bucket.dependencies) == 0
    assert len(bucket.internal_objects) == 1
    assert bucket.resource_type == test_bucket


def test_get_vm():
    """Test get vm model and its dependencies."""
    test_vm = 'test/vm'
    test_network = 'test/network'
    _destroy_dir = __setup_local({
        'network.json':
            """
{
    "dependencies": {},
    "internalObjects": {},
    "attributes":{
        "region": {
            "optional": true,
            "default": "euw",
            "cdk_key": "region",
            "var_type": "str"
        }
    },
    "cdk_name_key": "name",

    "cdk_provider":"test_provider",
    "cdk_module":"test_module",
    "cdk_class":"bucket_class"
}
""",
            'vm.json':
            """
{
    "dependencies": {
        "network": {
            "resource": "test/network",
            "default": {}
        }
    },
    "internalObjects": {},
    "attributes":{
        "region": {
            "optional": true,
            "default": "euw",
            "cdk_key": "region",
            "var_type": "str"
        },
        "type": {
            "optional": false,
            "cdk_key": "type"
        }
    },
    "cdk_name_key": "name",

    "cdk_provider":"test_provider",
    "cdk_module":"test_module",
    "cdk_class":"vm_class"
}
""",
    })
    resources = [test_vm]

    repo = LocalRepo(Path.cwd().as_posix())

    models = repo.get(resources)
    _destroy_dir()

    assert isinstance(models, dict)
    assert len(models) == 2

    assert test_vm in models
    assert isinstance(models[test_vm], ResourceModel)

    vm = models[test_vm]

    assert vm.resource_type == test_vm
    assert len(vm.attributes) == 2
    assert len(vm.dependencies) == 1

    assert test_network in models
    assert isinstance(models[test_network], ResourceModel)

    network = models[test_network]

    assert network.resource_type == test_network
    assert len(network.attributes) == 1
    assert len(network.dependencies) == 0


def test_cyclic_import():
    """Test erroneous cyclic import."""
    test_cyclic = 'test/cyclic'
    _destroy_dir = __setup_local({
        'cyclic.json':
            """
{
    "dependencies": {
        "network": {
            "resource": "test/cyclic",
            "default": {}
        }
    },
    "internalObjects": {},
    "attributes":{
        "region": {
            "optional": true,
            "default": "euw",
            "cdk_key": "region"
        }
    },
    "cdk_name_key": "name",

    "cdk_provider":"test_provider",
    "cdk_module":"test_module",
    "cdk_class":"vm_class"
}
""",
    })
    resources = [test_cyclic]
    repo = LocalRepo(Path.cwd().as_posix())

    models = repo.get(resources)
    _destroy_dir()

    assert isinstance(models, dict)
    assert len(models) == 1

    assert test_cyclic in models
    assert isinstance(models[test_cyclic], ResourceModel)


def test_model_not_found():
    """Test non existing model."""
    resources = ['not_an_existing_model']

    repo = LocalRepo('THipster/models')

    with pytest.raises(ModelNotFoundError):
        repo.get(resources)



"""parsedFile.py Module."""

from abc import ABC, abstractmethod


class ParsedValue(ABC):
    """Parsed Value Interface."""

    @property
    @abstractmethod
    def unmarshalled(self):
        """Return the unmarshalled value of the object."""
        pass


class Position:
    """Represents the position of a token.

    Indicates the initial position of a token, resource or character in the input files.
    It includes the file name, line and column numbers of the designated element.
    """

    def __init__(self, filename: str, ln: int, col: int):
        """Represent the position of a token.

        Indicates the initial position of a token, resource or character in the input
        files.
        It includes the file name, line and column numbers of the designated element.

        Parameters
        ----------
        fileName : str
            file name
        ln : int
            line number
        col : int
            column number
        """
        self.filename = filename
        self.ln = ln
        self.col = col

    def __str__(self) -> str:
        """Return a string of the position object.

        Returns
        -------
        str
            (File : {}, Ln {}, Col {})
        """
        return f'(File : {self.filename}, Ln {self.ln}, Col {self.col})'

    def __eq__(self, __value: object) -> bool:
        """Check if 2 positions are equal.

        Parameters
        ----------
        __value : object
            Position to compare

        Returns
        -------
        bool
            True if both positions are equal or False otherwise

        Raises
        ------
        TypeError
            If '__value' is not a Position
        """
        if not isinstance(__value, Position):
            msg = 'Value must be a Position'
            raise TypeError(msg)

        return (
            self.filename == __value.filename and
            self.ln == __value.ln and
            self.col == __value.col
        )


class ParsedAttribute:
    """Reprensents a Parsed Attribute Object."""

    def __init__(self, name: str, position: Position | None, value: ParsedValue):
        """
        Represent a Parsed Attribute Object.

        Parameters
        ----------
        name : str
            name of the attribute
        position : Position | None
            position of the attribute in its origin file
        value : ParsedValue
            value of the attribute
        """
        self.name: str = name
        self.position: Position | None = position
        self.__value = value

    @property
    def value(self):
        """Value of the parsed attribute."""
        return self.__value.value

    @property
    def unmarshalled_value(self):
        """Value of the parsed attribute."""
        return self.__value.unmarshalled


class ParsedList(ParsedValue):
    """Represents a Parsed List Object."""

    def __init__(self, value: list[ParsedValue]):
        """
        Represent a Parsed List Object.

        Parameters
        ----------
        value : list[ParsedValue]
            value of the parsed list
        """
        super().__init__()
        self.value: list[ParsedValue] = value

    @property
    def unmarshalled(self):
        """Return the unmarshalled value."""
        return [e.unmarshalled for e in self.value]

    def __iter__(self):
        """Return an iterator object."""
        self.i = 0
        return self

    def __next__(self):
        """Return the next value of the iterator."""
        if self.i > len(self.__value):
            raise StopIteration

        self.i += 1
        return self.value[self.i-1]


class ParsedLiteral(ParsedValue):
    """Represents a Parsed Literal Object."""

    def __init__(self, value: bool | int | float | str):
        """
        Represent a Parsed Literal Object.

        Parameters
        ----------
        value : Literal
            value of the parsed literal
        """
        super().__init__()
        self.value: bool | int | float | str = value

    @property
    def unmarshalled(self):
        """Return the unmarshalled value."""
        return self.value


class ParsedDict(ParsedValue):
    """Represents a Parsed Dictionnary Object."""

    def __init__(self, value: list[ParsedAttribute]):
        """
        Represent a Parsed Dictionnary Object.

        Parameters
        ----------
        value : list[ParsedAttribute]
            value of the parsed dictionnary
        """
        super().__init__()
        self.value: list[ParsedAttribute] = value

    @property
    def unmarshalled(self):
        """Return the unmarshalled value."""
        return {e.name: e.unmarshalled_value for e in self.value}


class ParsedResource:
    """Represents a Parsed Resource."""

    def __init__(
            self,
            parsed_resource_type: str,
            name: str,
            position: Position | None,
            attributes: list[ParsedAttribute],
    ):
        """
        Represent a Parsed Resource.

        Parameters
        ----------
        parsed_resource_type : str
        name : str
            name of the resource
        position : Position
            position of the resource in its origin file
        attributes : list[ParsedAttribute]
            list of attributes of the resource
        """
        self.resource_type: str = parsed_resource_type
        self.name: str = name
        self.position: Position | None = position
        self.attributes: list[ParsedAttribute] = attributes


class ParsedFile:
    """Represents a Parsed File.

    Object containing a list of parsed resources making up a file.
    """

    def __init__(self):
        """Represent a Parsed File."""
        self.resources: list[ParsedResource] = []



"""Test the AST of the DSL parser."""
import thipster.parser.dsl_parser.ast as ast
from thipster.engine.parsed_file import Position
from thipster.parser.dsl_parser.token import Token


def test_create_ast():
    """Test the creation of an AST."""
    # bucket nom-#test \n\t toto: tata
    tree = ast.FileNode()

    tree.add(
        ast.ResourceNode(
            resource_type=ast.StringNode(
                Token(Position('file', 1, 1), 'STRING', 'bucket'),
            ),
            name=ast.StringExprNode(
                Token(Position('file', 1, 8), 'STRING', 'nom-'),
                Token(Position('file', 1, 8), 'VAR', 'test'),
            ),

            # amount = IntNode(),

            # condition = IfNode(),

            parameters=ast.DictNode(
                [
                    ast.ParameterNode(
                        name=Token(Position('file', 2, 3), 'STRING', 'toto'),
                        value=ast.LiteralNode(
                            Token(Position('file', 2, 9), 'STRING', 'tata'),
                        ),

                        # amount = IntNode(),

                        # condition = IfElseNode(),
                    ),
                ],
            ),
        ),
    )

    assert str(tree) == '<RESOURCE \
type = <STRING (STRING bucket)>, \
name = <STRING-EXPR (STRING nom-) (VAR test)>, \
parameters = <DICT <PARAMETER name = (STRING toto), value = <LITERAL (STRING tata)>>>>'



"""Module containing the Token class and the TOKENTYPES enum."""
from enum import Enum

from thipster.engine.parsed_file import Position


class TOKENTYPES(Enum):
    """Contains all the possible token types."""

    AMOUNT = 'AMOUNT'
    AND = 'AND'
    BOOLEAN = 'BOOLEAN'
    BRACKETS_START = 'BRACKETS_START'
    BRACKETS_END = 'BRACKETS_END'
    COLON = 'COLON'
    COMMA = 'COMMA'
    ELIF = 'ELIF'
    ELSE = 'ELSE'
    EOF = 'EOF'
    EXCLAMATION = 'EXCLAMATION'
    FLOAT = 'FLOAT'
    IF = 'IF'
    INT = 'INT'
    NEWLINE = 'NEWLINE'
    NOT = 'NOT'
    OR = 'OR'
    PARENTHESES_START = 'PARENTHESES_START'
    PARENTHESES_END = 'PARENTHESES_END'
    STRING = 'STRING'
    TAB = 'TAB'
    VAR = 'VAR'
    WHITESPACE = 'WHITESPACE'

    PLUS = 'PLUS'
    MINUS = 'MINUS'
    MUL = 'MUL'
    DIV = 'DIV'
    PERCENT = 'PERCENT'
    POW = 'POW'
    EQ = 'EQ'
    EE = 'EE'
    NE = 'NE'
    LT = 'LT'
    LTE = 'LTE'
    GT = 'GT'
    GTE = 'GTE'

    def __str__(self) -> str:
        """Return the string representation of the token type."""
        return self.value


class Token:
    """Represents a Token of the THipster DSL."""

    def __init__(
        self, position: Position,
        token_type: TOKENTYPES, value: str | None = None,
    ):
        """
        Represent a Token.

        Parameters
        ----------
        position : Position
            Position of the token in its input file
        tokenType : str
            Token type
        value : str, optional
            Token value, by default None
        """
        self.position = position
        self.token_type = token_type
        self.value = value

    def __eq__(self, __value: object) -> bool:
        """Check if two tokens are equal."""
        if not isinstance(__value, Token):
            msg = 'Value must be a Token'
            raise TypeError(msg)

        return (
            self.position == __value.position and
            self.token_type == __value.token_type and
            self.value == __value.value
        )

    def __repr__(self) -> str:
        """Return the string representation of the token."""
        token_string = f'(Type: {self.token_type!s}, Position: {self.position!s}'
        if self.value:
            token_string += f', Value: {self.value}'
        token_string += ')'

        return token_string

    def __str__(self) -> str:
        """Return the string value of the token."""
        return f'({self.token_type} {self.value})'



"""Tests for the ResourceModel class."""
import thipster.engine.resource_model as rm


def test_create_empty_resource():
    """Test creating a resource with no attributes or dependencies."""
    model = rm.ResourceModel(
        'test_type',
        attributes=None,
        dependencies=None,
        internal_objects=None,
        name_key='test_key',
        cdk_provider='test_provider',
        cdk_module='test_module',
        cdk_name='test_name',
    )

    assert isinstance(model, rm.ResourceModel)
    assert model.resource_type == 'test_type'
    assert model.attributes is None
    assert model.dependencies is None
    assert model.internal_objects is None


def test_create_resource_with_dependencies():
    """Test creating a resource with dependencies."""
    dependencies = [
        rm.ResourceModel(
            'dependency_'+str(i),
            attributes=None,
            dependencies=None,
            internal_objects=None,
            name_key='test_key',
            cdk_provider='test_provider',
            cdk_module='test_module',
            cdk_name='test_name_'+str(i),
        ) for i in range(3)
    ]

    model = rm.ResourceModel(
        'test_type',
        attributes=None,
        dependencies=dependencies,
        internal_objects=None,
        name_key='test_key',
        cdk_provider='test_provider',
        cdk_module='test_module',
        cdk_name='test_name',
    )

    assert isinstance(model, rm.ResourceModel)
    assert model.resource_type == 'test_type'
    assert model.attributes is None
    assert isinstance(model.dependencies, list)
    for dep in model.dependencies:
        assert isinstance(dep, rm.ResourceModel)
        assert 'dependency_' in dep.resource_type


def test_create_num_attr():
    """Test creating a resource with a numeric attribute."""
    val = rm.ModelLiteral(3)

    attr = rm.ModelAttribute(
        cdk_name='test_attr',
        default=val, optional=False,
    )

    assert isinstance(attr, rm.ModelAttribute)
    assert isinstance(attr._ModelAttribute__default, rm.ModelLiteral)
    assert attr.default == 3


def test_create_str_attr():
    """Test creating a string attribute."""
    val = rm.ModelLiteral('test_default')

    attr = rm.ModelAttribute(
        cdk_name='test_attr',
        default=val, optional=False,
    )

    assert isinstance(attr, rm.ModelAttribute)
    assert isinstance(attr._ModelAttribute__default, rm.ModelLiteral)
    assert attr.default == 'test_default'


def test_create_list_str_attr():
    """Test creating a list of strings attribute."""
    val = rm.ModelList(
        [rm.ModelLiteral('test_default_' + str(i)) for i in range(3)],
    )

    attr = rm.ModelAttribute(
        cdk_name='test_attr',
        default=val, optional=False,
    )

    assert isinstance(attr, rm.ModelAttribute)
    assert isinstance(attr._ModelAttribute__default, rm.ModelList)

    assert len(attr.default) == 3
    for val in attr.default:
        assert 'test_default' in val.value


def test_create_dict_str_attr():
    """Test creating a dict of strings attribute."""
    val = rm.ModelDict(
        {rm.ModelLiteral('test_default_' + str(i)) for i in range(3)},
    )

    attr = rm.ModelAttribute(
        cdk_name='test_attr',
        default=val, optional=False,
    )

    assert isinstance(attr, rm.ModelAttribute)
    assert isinstance(attr._ModelAttribute__default, rm.ModelDict)

    assert len(attr.default) == 3
    for val in attr.default:
        assert 'test_default' in val.value


def test_create_resource_with_attributes():
    """Test creating a resource with attributes."""
    attr = {
        'attr_num': rm.ModelAttribute(
            cdk_name='attr_num',
            default=rm.ModelLiteral(3),
            optional=False,
        ),
        'attr_str': rm.ModelAttribute(
            cdk_name='attr_str',
            default=rm.ModelLiteral('test'),
            optional=False,
        ),
        'attr_list': rm.ModelAttribute(
            cdk_name='attr_list',
            default=rm.ModelList([
                rm.ModelLiteral('test_list_1'),
                rm.ModelLiteral('test_list_2'),
                rm.ModelLiteral('test_list_3'),
            ]),
            optional=False,
        ),
        'attr_dict': rm.ModelAttribute(
            cdk_name='attr_dict',
            default=rm.ModelDict([
                rm.ModelAttribute(
                    'attr_dict_1',
                    default=rm.ModelLiteral('test'),
                    optional=False,
                ),
                rm.ModelAttribute(
                    'attr_dict_2',
                    default=rm.ModelLiteral('test'),
                    optional=False,
                ),
            ]),
            optional=False,
        ),
    }

    model = rm.ResourceModel(
        'test_type',
        attributes=attr,
        dependencies=None,
        internal_objects=None,
        name_key='test_key',
        cdk_provider='test_provider',
        cdk_module='test_provider',
        cdk_name='test_provider',
    )

    assert isinstance(model, rm.ResourceModel)
    assert model.resource_type == 'test_type'
    assert isinstance(model.attributes, dict)
    for _, attribute in model.attributes.items():
        assert isinstance(attribute, rm.ModelAttribute)
        assert isinstance(
            attribute._ModelAttribute__default, rm.ModelValue,
        )
        if isinstance(attribute._ModelAttribute__default, rm.ModelDict):
            for val in attribute.default:
                assert isinstance(val, rm.ModelAttribute)

        elif isinstance(attribute._ModelAttribute__default, rm.ModelList):
            for val in attribute.default:
                assert isinstance(val, rm.ModelValue)

    assert model.dependencies is None


def test_create_internal_object():
    """Test creating an internal object."""
    model = rm.ResourceModel(
        'test_type',
        attributes=None,
        dependencies=None,
        internal_objects=None,
        name_key=None,
        cdk_provider='test_provider',
        cdk_module='test_module',
        cdk_name='test_name',
    )

    assert isinstance(model, rm.ResourceModel)
    assert model.name_key is None



"""JSONRepo.py module."""

import json
from abc import ABC, abstractmethod

import thipster.engine.resource_model as rm
from thipster.engine import RepositoryPort


class JSONRepo(RepositoryPort, ABC):
    """Represents a JSON resources Repository.

    JSON Models of resources and services offered by supported cloud providers are
    stored in a repository.
    """

    _parent_stack = []

    def __init__(self) -> None:
        """Represent a JSON resources Repository.

        JSON Models of resources and services offered by supported cloud providers are
        stored in a repository.
        """
        super().__init__()
        self.model_list = {}

    @abstractmethod
    def get_json(self, name: str) -> str | bytes | bytearray:
        """Get the JSON file corresponding to the name given."""
        raise Exception

    def get(self, resource_names: list[str]) -> dict[str, rm.ResourceModel]:
        """Get the corresponding resource Models from a list of names.

        Parameters
        ----------
        resourceNames : list[str]
            List of the desired resource models names

        Returns
        -------
        dict[str, rm.ResourceModel]
            Dictionnary of the corresponding resource models
        """
        for resource in resource_names:
            self.__add_model(resource)

        return self.model_list

    def __create_value(self, val: object | None) -> rm.ModelValue | None:
        """Create the right Model value implementation from the raw JSON.

        Parameters
        ----------
        val : object
            Raw JSON model's attribute value

        Returns
        -------
        ModelValue | None
            Value of the attribute, implementation of ModelValue : ModelDict,
            ModelList, ModelLiteral or None
        """
        if val is None:
            return None
        if isinstance(val, dict):
            return rm.ModelDict(None)
        if isinstance(val, list):
            return rm.ModelList([self.__create_value(i) for i in val])

        return rm.ModelLiteral(val)

    def __create_attribute(self, raw: dict[str, str]) -> list[rm.ModelAttribute]:
        """Create a model's attributes from the raw JSON input.

        Parameters
        ----------
        raw : dict[str, str]
            Json model's raw attributes

        Returns
        -------
        list[ModelAttribute]
            Attributes of the resource model
        """
        attributes = {}

        for name, attr in raw.items():
            optional = attr['optional'] if 'optional' in attr else True

            value = attr['default'] if 'default' in attr else None

            is_list = 'list' in attr['var_type'] if 'var_type' in attr else False

            default = self.__create_value(value)

            attributes[name] = rm.ModelAttribute(
                attr['cdk_key'],
                default=default,
                optional=optional,
                is_list=is_list,
            )

        return attributes

    def __create_model(self, name: str) -> rm.ResourceModel:
        """Get the json file and create a Resource Model.

        Parameters
        ----------
        name : str
            Name of the model to find

        Returns
        -------
        ResourceModel
            Resource model corresponding to the name given
        """
        model = self.get_json(name)

        json_model = json.loads(model)

        for _, dep in json_model['dependencies'].items():
            if dep['resource'] not in self.model_list.keys():
                self.__add_model(dep['resource'])

        for _, internal_object in json_model['internalObjects'].items():
            if internal_object['resource'] not in self.model_list.keys():
                self.__add_model(internal_object['resource'])

        return rm.ResourceModel(
            name,
            attributes=self.__create_attribute(json_model['attributes']),
            dependencies=json_model['dependencies'],
            internal_objects=json_model['internalObjects'],
            name_key=json_model['cdk_name_key']
            if 'cdk_name_key' in json_model else None,
            cdk_provider=json_model['cdk_provider'],
            cdk_module=json_model['cdk_module'],
            cdk_name=json_model['cdk_class'],
        )

    def __add_model(self, model: str) -> rm.ResourceModel:
        """Add a model to the list.

        Parameters
        ----------
        model : str
            Name of the model to add

        Returns
        -------
        ResourceModel
            Resource model corresponding to the name given
        """
        if model not in self.model_list.keys():
            self.model_list[model] = None
            self.model_list[model] = self.__create_model(model)

        return self.model_list[model]



"""Repository module.

Contains the various options to store the JSON models used by THipster to generate the
Python Terraform CDK constructs.
Currently, two options are available:
- A local repository, which stores the models in a local folder
- A GitHub repository
"""

from .github import GithubRepo
from .local import LocalRepo



"""Helper functions for thipster."""
import logging


def create_logger(class_name: str) -> logging.Logger:
    """Create a logger for the given class."""
    return logging.getLogger(class_name)


def set_logging_config(
    log_level: str = 'INFO',
    filename: str | None = None,
    filemode: str = 'w',
) -> None:
    """Set logging configuration.

    Parameters
    ----------
    log_level : str
        Log level, defaults to INFO
    filename : str, optional
        Log filename, defaults to None

    """
    logging.basicConfig(
        level=log_level,
        format='%(asctime)s %(levelname)s %(name)s %(message)s',
        filename=filename,
        filemode=filemode,
        encoding='utf-8',
    )



"""Parser module interface."""
from abc import ABC, abstractclassmethod

from thipster.engine.parsed_file import ParsedFile


class ParserPort(ABC):
    """Parser port."""

    @classmethod
    @abstractclassmethod
    def run(
        cls,
        path: str,  # noqa: ARG003
    ) -> ParsedFile:
        """Abstract run method.

        Parameters
        ----------
        path : str
            The path of the filesor directory to be processed

        Returns
        -------
        ParsedFile
            Object containing a representation of an input file broken down into
            individual resources

        Raises
        ------
        NotImplementedError
            If method is not implemented in inheriting classes

        """
        msg = 'Should implement run()'
        raise NotImplementedError(msg)



"""Local repository implementation."""
from pathlib import Path

from .exceptions import ModelNotFoundError
from .json import JSONRepo


class LocalRepo(JSONRepo):
    """Represents a local resources Repository.

    JSON Models of resources and services offered by supported cloud providers are
    stored in a repository.
    This class is used to access those models if they are stored locally.
    """

    def __init__(self, path: str) -> None:
        """Represent a local resources Repository.

        JSON Models of resources and services offered by supported cloud providers are
        stored in a repository.
        This class is used to access those models if they are stored locally.
        """
        super().__init__()
        self.__repo = path

    def get_json(self, name: str) -> str | bytes | bytearray:
        """Get the json file from the local repository."""
        name += '.json'
        try:
            with (Path(self.__repo)/name).open() as file:
                return file.read()
        except Exception:
            raise ModelNotFoundError(name)



"""Repository module interface."""
from abc import ABC, abstractmethod

from thipster.engine.resource_model import ResourceModel


class RepositoryPort(ABC):
    """Repository port."""

    @abstractmethod
    def get(self, resource_names: list[str]) -> dict[str, ResourceModel]:
        """Abstract get method.

        Parameters
        ----------
        resourceNames : list[str]
            List of resource names to be retrieved from the repository

        Returns
        -------
        dict[str, ResourceModel]
            Dictionary of ResourceModel objects, indexed by resource name
        """
        pass



"""Central engine module for THipster.

This module contains the Engine class, which is the central class of the THipster
package as well as the interfaces for the various components of the application.
"""

from .engine import Engine
from .exceptions import THipsterError
from .i_auth import AuthPort
from .i_parser import ParserPort
from .i_repository import RepositoryPort
from .i_terraform import TerraformPort



"""Parser module.

Contains the various parsers used by THipster to interpret the input files.
Currently, two options are available:
- The DSL parser, which interprets .thips files
- The YAML JINJA parser, which interprets .yml/.yaml/.jinja files
"""

from .dsl_parser import DSLParser
from .parser_factory import ParserFactory
from .yaml_parser import YAMLParser



"""THipster DSL token parser module."""
import thipster.parser.dsl_parser.ast as ast

from .exceptions import (
    DSLConditionError,
    DSLSyntaxError,
    DSLUnexpectedEOFError,
)
from .token import TOKENTYPES as TT
from .token import Token


class TokenParser:
    """Parse the tokens into an AST (Abstract Syntax Tree)."""

    def __init__(self, tokens: list[Token]) -> None:
        """Parse the tokens into an AST (Abstract Syntax Tree).

        Parameters
        ----------
        tokens : list[Token]
            The list of tokens to parse.
        """
        self.__tokens = tokens

    def run(self) -> ast.FileNode:
        """Run the parser."""
        self.__rm_empty_lines()
        tree = ast.FileNode()
        try:
            self.__trim_newlines()
            while self.__get_next_type() != TT.EOF:
                self.__trim_newlines()

                tree.add(self.__create_resource())
                self.__trim_newlines()
        except Exception as e:
            raise e

        return tree

    def __next(self, expected: TT | list[TT] | None = None) -> Token:
        """Get next token and pop it from the list.

        Parameters
        ----------
        expected : TT | list[TT] | None
            Expected token type(s), by default None

        Returns
        -------
        Token
            The next token
        """
        next_token_type = self.__get_next_type()

        if expected:
            if type(expected) is list:
                if next_token_type not in expected:
                    raise DSLSyntaxError(self.__tokens[0], expected)
            elif next_token_type != expected:
                raise DSLSyntaxError(self.__tokens[0], expected)
        return self.__tokens.pop(0)

    def __check(self, expected: TT, index: int = 0) -> Token | None:
        """Check if the next token is the expected one.

        If the type of the next token is equal to the expected parameter, it is popped
        from the list.

        Parameters
        ----------
        expected : TT
            Expected token type
        index : int, optional
            Index of the token to check, by default 0

        Returns
        -------
        Token | None
            The next token if it is the expected one, None otherwise
        """
        token = self.__get_next_type(index=index)
        if token != expected:
            return None

        return self.__next(expected)

    def __get_next_type(self, index: int = 0) -> TT:
        """Get the type of the next token.

        Parameters
        ----------
        index : int, optional
            Index of the token to get, by default 0

        Returns
        -------
        TT
            The type of the next token
        """
        if len(self.__tokens) <= index:
            raise DSLUnexpectedEOFError

        return self.__tokens[index].token_type

    def __trim_newlines(self):
        while self.__check(TT.NEWLINE):
            pass

    def __rm_empty_lines(self):
        # Detect empty line
        empty_types = [TT.TAB, TT.WHITESPACE]
        end = 0
        while end < len(self.__tokens):
            begin = end
            while self.__get_next_type(end) in empty_types:
                end += 1

            if self.__get_next_type(index=end) == TT.NEWLINE:
                for _ in range(begin, end):
                    self.__tokens.pop(begin)
            end += 1

    def __get_newline(self):
        newline = self.__next(TT.NEWLINE)
        self.__trim_newlines()

        return newline

    def __get_whitespaces(self):
        while self.__check(TT.WHITESPACE):
            pass

    def __get_tabs(self, indent: int) -> bool:
        """Check if the number of tabs is correct/ if it is the end of the block."""
        if self.__get_next_type() == TT.EOF:
            return False
        if self.__get_next_type(indent-1) == TT.TAB:
            for _ in range(indent):
                self.__next(TT.TAB)
            return True
        return False

    def __create_resource(
        self,
        indent=0,
    ) -> ast.ResourceNode | ast.IfNode | ast.AmountNode:
        r"""Create an AST Resouce, If or Amount node.

        Format: type, name, ":", [amt_ctrl], [if_ctrl] ,"\\n"
        (list | dict | {parameter, "\\n"}).
        """
        try:
            for _ in range(indent):
                self.__next(TT.TAB)

            resource_type = self.__get_type()
            self.__get_whitespaces()
            name = self.__get_string_expr()
            self.__get_whitespaces()
            self.__next(TT.COLON)
            self.__get_whitespaces()

            if_ctrl = self.__get_if_ctrl()
            self.__get_whitespaces()
            nb_ctrl = self.__get_nb_ctrl()
            self.__get_whitespaces()

            self.__get_newline()

            properties = self.__get_properties(indent+1)

        except DSLSyntaxError as e:
            raise e

        resource = ast.ResourceNode(
            resource_type=resource_type,
            name=name,
            parameters=properties,
        )

        if if_ctrl:
            if_ctrl.if_case = resource
            resource = if_ctrl

        if nb_ctrl:
            nb_ctrl.node = resource
            resource = nb_ctrl

        return resource

    def __get_parameter(self, indent: int) -> ast.ParameterNode:
        r"""Create an AST Parameter node.

        Format: name, ":", (value, [if_else_ctrl] | [if_ctrl], "\\n", (list | dict)).
        """
        try:
            name = self.__next(TT.STRING)
            self.__get_whitespaces()
            self.__next(TT.COLON)
            self.__get_whitespaces()
        except DSLSyntaxError as e:
            raise e

        next_token_type = self.__get_next_type()

        if next_token_type not in [
            TT.IF,
            TT.NEWLINE,
        ]:
            # value, [if_else_ctrl]
            try:
                value = self.__get_value()
                self.__get_whitespaces()
                if_else_ctrl = self.__get_if_else_ctrl()
                self.__get_whitespaces()
            except:
                raise

            parameter = ast.ParameterNode(
                name=ast.StringNode(name),
                value=value,
            )

            if if_else_ctrl:
                if_else_ctrl.if_case = parameter.value
                parameter.value = if_else_ctrl

            return parameter

        # [if_ctrl], "\n" (liste | dict)
        try:
            if_ctrl = self.__get_if_ctrl()
            self.__get_whitespaces()
            newline = self.__get_newline()
            properties = self.__get_properties(indent+1)
        except DSLSyntaxError as e:
            if e.token.token_type == TT.TAB:
                raise DSLSyntaxError(
                    token=newline,
                    expected=TT.STRING,
                )

            raise e

        parameter = ast.ParameterNode(
            name=ast.StringNode(name),
            value=properties,
        )

        if if_ctrl:
            if_ctrl.if_case = parameter
            parameter = if_ctrl

        return parameter

    def __get_properties(self, indent: int) -> list[ast.ParameterNode]:
        """Create a list of AST Parameter nodes.

        Format: (list | dict).
        """
        i = indent
        try:
            next_token_type = self.__get_next_type(indent)
            while next_token_type == TT.WHITESPACE:
                i += 1
                next_token_type = self.__get_next_type(i)

            if next_token_type == TT.STRING:
                props = self.__get_dict(indent)

            elif next_token_type == TT.MINUS:
                props = self.__get_list(indent)

            else:
                raise DSLSyntaxError(self.__next(), TT.TAB)

        except DSLUnexpectedEOFError as eof:
            if indent > 1:
                raise DSLUnexpectedEOFError from eof

            if eof.__cause__:
                raise DSLUnexpectedEOFError from eof.__cause__

            props = ast.DictNode([])

        return props

    def __get_list(self, indent: int) -> ast.ListNode:
        r"""Create an AST List node.

        Format: { "-", value, [amt_ctrl], [if_else_ctrl], "\\n"}.
        """
        list_items = []

        try:
            while self.__get_tabs(indent):
                self.__check(TT.MINUS)
                self.__get_whitespaces()

                value = self.__get_value()
                self.__get_whitespaces()

                if_else_ctrl = self.__get_if_else_ctrl()
                self.__get_whitespaces()
                amount_ctrl = self.__get_nb_ctrl()
                self.__get_whitespaces()

                if if_else_ctrl:
                    if_else_ctrl.if_case = value
                    value = if_else_ctrl

                if amount_ctrl:
                    amount_ctrl.node = value
                    value = amount_ctrl

                list_items.append(value)

                self.__get_newline()
        except Exception as e:
            raise e

        self.__tokens.insert(
            0, Token(
                position=self.__tokens[0].position,
                token_type=TT.NEWLINE,
            ),
        )

        return ast.ListNode(list_items)

    def __get_inline_list(self) -> ast.ListNode:
        list_items = []

        self.__next(TT.BRACKETS_START)
        self.__get_whitespaces()
        next_token_type = self.__get_next_type()

        if next_token_type in [
            TT.BOOLEAN,
            TT.FLOAT,
            TT.INT,
            TT.STRING,
            TT.VAR,
            TT.BRACKETS_START,
        ]:
            list_items.append(self.__get_value())
            while not self.__check(TT.BRACKETS_END):
                self.__next([TT.COMMA, TT.NEWLINE])
                while self.__check(TT.TAB):
                    pass
                self.__get_whitespaces()

                list_items.append(self.__get_value())
                self.__get_whitespaces()
        else:
            self.__next(TT.BRACKETS_END)

        return ast.ListNode(list_items)

    def __get_dict(self, indent: int) -> ast.DictNode:
        r"""Create an AST Dict node.

        Format: { parameter, "\\n" }.
        """
        parameters = []

        try:
            while self.__get_tabs(indent):
                self.__get_whitespaces()
                parameters.append(self.__get_parameter(indent))
                self.__get_whitespaces()
                self.__get_newline()
        except Exception as e:
            raise e

        self.__tokens.insert(
            0, Token(
                position=self.__tokens[0].position,
                token_type=TT.NEWLINE,
            ),
        )

        return ast.DictNode(parameters)

    def __get_nb_ctrl(self) -> ast.AmountNode | None:
        """Create an AST Amount node.

        Format: "amount", ":", int, ["#", var].
        """
        try:
            amount_token = self.__check(TT.AMOUNT)
            if not amount_token:
                return None

            self.__get_whitespaces()
            self.__next(TT.COLON)

            self.__get_whitespaces()
            amount = self.__get_value()

            self.__get_whitespaces()
        except DSLSyntaxError as e:
            raise e

        amount_variable = self.__check(TT.VAR)

        return ast.AmountNode(
            position=amount_token,
            amount=amount,
            variable=ast.VariableDefinitionNode(
                amount_variable, ast.IntNode(Token(None, TT.INT, 1)),
            )
            if amount_variable else None,
            node=None,
        )

    def __get_if_ctrl(self) -> ast.IfNode | None:
        """Create an AST If node.

        Format: "if", condition.
        """
        condition = self.__check(TT.IF)
        if not condition:
            return None

        self.__get_whitespaces()
        try:
            condition = self.__get_comp_expr()
            self.__get_whitespaces()
        except DSLSyntaxError as e:
            raise e

        return ast.IfNode(
            condition=condition,
            if_case=None,
        )

    def __get_if_else_ctrl(self) -> ast.IfElseNode | None:
        """Create an AST IfElse node.

        Format: if_ctrl, ["else" , value].
        """
        if_ctrl = self.__get_if_ctrl()
        self.__get_whitespaces()
        if not if_ctrl:
            return None

        if self.__check(TT.ELSE):
            self.__get_whitespaces()
            else_case = self.__get_value()
            self.__get_whitespaces()
        else:
            else_case = None

        return ast.IfElseNode(
            condition=if_ctrl.condition,
            if_case=None,
            else_case=else_case,
        )

    def __get_value(self) -> ast.LiteralNode:
        next_token_type = self.__get_next_type()
        match next_token_type:
            case TT.STRING:
                value = self.__get_string_expr()

            case TT.VAR:
                value = ast.LiteralNode(ast.VariableNode(self.__next(TT.VAR)))

            case TT.BRACKETS_START:
                value = self.__get_inline_list()

            case TT.NOT | TT.BOOLEAN:
                value = self.__get_comp_expr()

            case _:
                value = self.__get_arith_expr()
                self.__get_whitespaces()

                next_token_type = self.__get_next_type()
                if next_token_type in [TT.EE, TT.LT, TT.GT, TT.LTE, TT.GTE]:
                    operator = self.__next(
                        [TT.EE, TT.LT, TT.GT, TT.LTE, TT.GTE],
                    )
                    self.__get_whitespaces()
                    right_expression = self.__get_comp_expr()
                    self.__get_whitespaces()
                    self.__next(TT.PARENTHESES_END)
                    self.__get_whitespaces()

                    value = ast.CompExprNode(value, operator, right_expression)

        if_else_ctrl = self.__get_if_else_ctrl()
        if not if_else_ctrl:
            return value

        if_else_ctrl.if_case = value
        if if_else_ctrl.else_case is None:
            raise DSLSyntaxError([TT.STRING, TT.VAR])
        return if_else_ctrl

    def __get_comp_expr(self) -> ast.CompExprNode:
        try:
            next_token_type = self.__get_next_type()
            match next_token_type:

                case TT.NOT:
                    operator = self.__next(TT.NOT)
                    self.__get_whitespaces()
                    expression = self.__get_comp_expr()
                    self.__get_whitespaces()
                    return ast.CompExprNode(expression, operator)

                case TT.PARENTHESES_START:
                    self.__next(TT.PARENTHESES_START)
                    self.__get_whitespaces()
                    left_expression = self.__get_comp_expr()
                    self.__get_whitespaces()
                    operator = self.__next([TT.OR, TT.AND])
                    self.__get_whitespaces()
                    right_expression = self.__get_comp_expr()
                    self.__get_whitespaces()
                    self.__next(TT.PARENTHESES_END)
                    self.__get_whitespaces()

                    return ast.CompExprNode(left_expression, operator, right_expression)

                case TT.BOOLEAN:
                    return ast.LiteralNode(ast.BoolNode(self.__next(TT.BOOLEAN)))

                case _:
                    left_expression = self.__get_arith_expr()
                    self.__get_whitespaces()
                    operator = self.__next(
                        [TT.EQ, TT.LT, TT.GT, TT.EXCLAMATION],
                    )

                    # ==
                    if operator.token_type is TT.EQ:
                        self.__next(TT.EQ)
                        operator = Token(operator.position, TT.EE)

                    # !=
                    if operator.token_type is TT.EXCLAMATION:
                        self.__next(TT.EQ)
                        operator = Token(operator.position, TT.NE)

                    next_token_type = self.__get_next_type()
                    if next_token_type is TT.EQ:
                        self.__next(TT.EQ)
                        match operator.token_type:
                            case TT.LT:
                                operator = Token(operator.position, TT.LTE)
                            case TT.GT:
                                operator = Token(operator.position, TT.GTE)

                    self.__get_whitespaces()
                    right_expression = self.__get_arith_expr()
                    self.__get_whitespaces()
                    return ast.CompExprNode(left_expression, operator, right_expression)

        except DSLSyntaxError as e:
            raise DSLConditionError(e.token)

    def __get_arith_expr(self) -> ast.ArithExprNode:
        terms = []
        operator = []
        terms.append(self.__get_term())
        self.__get_whitespaces()

        next_token_type = self.__get_next_type()
        while next_token_type in [TT.PLUS, TT.MINUS]:
            operator.append(next_token_type)
            terms.append(self.__get_term())
            self.__get_whitespaces()
            next_token_type = self.__get_next_type()

        return ast.ArithExprNode(terms, operator)

    def __get_term(self) -> ast.TermNode:
        factors = []
        operator = []
        factors.append(self.__get_factor())
        self.__get_whitespaces()

        next_token_type = self.__get_next_type()
        while next_token_type in [TT.MUL, TT.DIV, TT.PERCENT]:
            operator.append(
                self.__next(
                    [TT.MUL, TT.DIV, TT.PERCENT],
                ).token_type,
            )
            self.__get_whitespaces()
            factors.append(self.__get_factor())
            self.__get_whitespaces()
            next_token_type = self.__get_next_type()

        return ast.TermNode(factors, operator)

    def __get_factor(self) -> ast.FactorNode:
        next_token_type = self.__get_next_type()

        match next_token_type:
            case TT.PLUS | TT.MINUS:
                operator = self.__next([TT.PLUS, TT.MINUS]).token_type
                self.__get_whitespaces()
                factor = self.__get_factor()
                self.__get_whitespaces()
                return ast.FactorNode([factor], operator)

            case _:
                factors = [self.__get_atom()]
                self.__get_whitespaces()

                next_token_type = self.__get_next_type()
                operator = TT.PLUS
                while next_token_type == TT.POW:
                    operator = self.__next(TT.POW).token_type
                    self.__get_whitespaces()
                    factors.append(self.__get_factor())
                    self.__get_whitespaces()
                    next_token_type = self.__get_next_type()
                return ast.FactorNode(factors, operator)

    def __get_atom(self):
        next_token_type = self.__get_next_type()
        match next_token_type:
            case TT.INT:
                return ast.IntNode(self.__next())

            case TT.VAR:
                return ast.VariableNode(self.__next())

            case TT.FLOAT:
                return ast.FloatNode(self.__next())

            case TT.PARENTHESES_START:
                self.__next(TT.PARENTHESES_START)
                self.__get_whitespaces()
                expression = self.__get_arith_expr()
                self.__get_whitespaces()
                self.__next(TT.PARENTHESES_END)
                self.__get_whitespaces()
                return expression

            case _:
                raise DSLSyntaxError(
                    self.__tokens[0], [TT.INT, TT.FLOAT, TT.PARENTHESES_START],
                )

    def __get_string_expr(self):
        values = []
        token = self.__next([TT.STRING, TT.VAR])
        match token.token_type:
            case TT.STRING:
                values.append(ast.StringNode(token))

            case TT.VAR:
                values.append(ast.VariableNode(token))

        next_token_type = self.__get_next_type()
        while next_token_type in [TT.STRING, TT.VAR, TT.INT, TT.MINUS, TT.DIV]:
            match next_token_type:
                case TT.STRING:
                    values.append(ast.StringNode(self.__next(TT.STRING)))

                case TT.VAR:
                    values.append(ast.VariableNode(self.__next(TT.VAR)))

                case TT.INT:
                    values.append(ast.IntNode(self.__next(TT.INT)))

                case TT.DIV:
                    token = Token(self.__next(TT.DIV).position, TT.STRING, '/')
                    values.append(ast.StringNode(token))

                case TT.MINUS:
                    token = Token(
                        self.__next(
                            TT.MINUS,
                        ).position, TT.STRING, '-',
                    )
                    values.append(ast.StringNode(token))
            next_token_type = self.__get_next_type()

        return ast.StringExprNode(values)

    def __get_type(self):
        values = []
        values.append(ast.StringNode(self.__next(TT.STRING)))

        next_token_type = self.__get_next_type()
        while next_token_type in [TT.STRING, TT.MINUS, TT.DIV]:
            match next_token_type:
                case TT.STRING:
                    values.append(ast.StringNode(self.__next(TT.STRING)))

                case TT.DIV:
                    token = Token(self.__next(TT.DIV).position, TT.STRING, '/')
                    values.append(ast.StringNode(token))

                case TT.MINUS:
                    token = Token(
                        self.__next(
                            TT.MINUS,
                        ).position, TT.STRING, '-',
                    )
                    values.append(ast.StringNode(token))
            next_token_type = self.__get_next_type()

        return ast.StringExprNode(values)



"""Setup file for the THipster-cli package."""
from pathlib import Path

from setuptools import find_packages, setup


def get_extra_requires() -> dict[str, list[str]]:
    """Get the extra requirements from the requirements folder."""
    extras_require = {}
    for req_file in Path('requirements').glob('requirements-*.txt'):
        extras_require[
            req_file.stem.removeprefix('requirements-')
        ] = req_file.read_text().splitlines()
    return extras_require


__version__ = '0.20.1'

with Path('requirements.txt').open() as f:
    required = f.read().splitlines()

with Path('README.md').open() as rm:
    readme = rm.read()

setup(
    name='thipster',
    authors=[
        {'name': 'rcattin', 'email': 'rafa.cattin+thipster@gmail.com'},
        {'name': 'gsuquet', 'email': 'gsuquet@ippon.fr'},
    ],
    description='THipster is a tool dedicated to simplifying the difficulty associated \
with writing Terraform files. It allows users to write infrastructure as code in a \
simplified format, using either YAML (with JINJA) or the dedicated Thipster DSL.',
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords=[
        'thipster',
        'terraform',
        'infrastructure',
        'infrastructure-as-code',
        'iac',
        'generator',
        'dsl',
        'yaml',
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],

    download_url='https://github.com/THipster/THipster.git',
    url='https://github.com/THipster/THipster',

    version=__version__,
    install_requires=required,
    packages=find_packages(
        exclude=['pipelines'],
    ),
    extras_require=get_extra_requires(),
)



"""Exceptions for the Terraform CDK module."""
from thipster.engine import THipsterError


class CDKError(THipsterError):
    """Shared base Exception for the Terraform CDK module."""

    pass


class CDKInvalidAttributeError(CDKError):
    """Exception raised when an attribute is invalid."""

    def __init__(self, attr: str, model_type: str, **args: object) -> None:
        """Exception raised when an attribute is invalid.

        Parameters
        ----------
        attr: str
            The attribute that is invalid
        model_type: str
            The model type that contains the attribute
        """
        super().__init__(*args)
        self.__attr = attr
        self.__modelType = model_type

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'{self.__attr} in {self.__modelType} but not useful'


class CDKMissingAttributeError(CDKError):
    """Exception raised when an attribute is missing."""

    def __init__(self, resource: str, attributes: list[str], **args: object) -> None:
        """Exception raised when an attribute is missing."""
        super().__init__(**args)
        self.__resource = resource
        self.__attributes = attributes

    @property
    def message(self) -> str:
        """Return the exception message."""
        attr_str = ', '.join(self.__attributes)
        return f'Missing {attr_str} in {self.__resource}'


class CDKMissingAttributeInDependencyError(CDKMissingAttributeError):
    """Exception raised when an attribute is missing in a dependency."""

    def __init__(self, resource: str, attributes: list[str], **args: object) -> None:
        """Exception raised when an attribute is missing in a dependency.

        Parameters
        ----------
        resource: str
            The resource where the attribute(s) is missing
        attributes: list[str]
            list of attributes that are missing
        """
        super().__init__(resource, attributes, **args)


class CDKDependencyNotDeclaredError(CDKError):
    """Exception raised when a dependency is used but not declared."""

    def __init__(
            self, dependency_type: str, dependency_name: str, **args: object,
    ) -> None:
        """Exception raised when a dependency is used but not declared.

        Parameters
        ----------
        dependency_type: str
            The type of the dependency
        dependency_name: str
            The name of the dependency
        """
        super().__init__(*args)
        self.__name = dependency_name
        self.__type = dependency_type

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'{self.__type} {self.__name} not declared \
(be sure to declare it before using it)'


class CDKCyclicDependenciesError(CDKError):
    """Exception raised when cyclic dependencies are detected."""

    def __init__(self, stack: list[str], **args: object) -> None:
        """Exception raised when cyclic dependencies are detected.

        Parameters
        ----------
        stack: list[str]
            The stack of dependencies that are cyclic
        """
        super().__init__(*args)
        self.__stack = stack

    @property
    def message(self) -> str:
        """Return the exception message."""
        return ','.join(self.__stack)



"""Exceptions for the THipster parser module."""
from thipster.engine import THipsterError


class ParserPathNotFoundError(THipsterError):
    """Exception raised when the parser cannot find the path."""

    def __init__(self, path: str, *args: object) -> None:
        """Exception raised when the parser cannot find the path.

        Parameters
        ----------
        path : str
            The path that was not found.
        """
        super().__init__(*args)

        self.__path = path

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'Path not found : {self.__path}'


class NoFileFoundError(THipsterError):
    """Exception raised when the parser cannot find any files to parse."""

    def __init__(self, path: str, *args: object) -> None:
        """Exception raised when the parser cannot find any files to parse.

        Parameters
        ----------
        path : str
            The path where no files were found.
        """
        super().__init__(*args)

        self.__path = path

    @property
    def message(self) -> str:
        """Return the exception message."""
        return f'No files to parse in {self.__path}'
